from conceptnet_rocks.conceptnet5.uri import uri_prefixes


def gin_indexable_edge(edge):
    """
    Convert an edge into a dictionary that can be matched with the JSONB @>
    operator, which tests if one dictionary includes all the information in
    another. This operator can be indexed by GIN.

    We replace the 'start', 'end', 'rel', and 'dataset' URIs with lists
    of their URI prefixes. We query those slots with a single-element list,
    which will be a sub-list of the prefix list if it's a match.

    As an example, a query for {'start': '/c/en'} will become the GIN
    query {'start': ['/c/en']}, which will match indexed edges such as
    {
        'start': ['/c/en', '/c/en/dog'],
        'end': ['/c/en', '/c/en/bark'],
        'rel': ['/r/CapableOf'],
        ...
    }
    """
    gin_edge = {}
    gin_edge['uri'] = edge['uri']
    gin_edge['start'] = uri_prefixes(edge['start'])
    gin_edge['end'] = uri_prefixes(edge['end'])
    gin_edge['rel'] = uri_prefixes(edge['rel'])
    gin_edge['dataset'] = uri_prefixes(edge['dataset'])
    flat_sources = set()
    for source in edge['sources']:
        for value in source.values():
            flat_sources.update(uri_prefixes(value, min_pieces=3))
    gin_edge['sources'] = sorted(flat_sources)
    return gin_edge



# These relations are symmetric: it doesn't matter which concept is their
# 'start' or 'end'.
SYMMETRIC_RELATIONS = {
    '/r/RelatedTo',
    '/r/SimilarTo',
    '/r/EtymologicallyRelatedTo',
    '/r/Synonym',
    '/r/Antonym',
    '/r/DistinctFrom',
}


from conceptnet_rocks.conceptnet5.nodes import ld_node
from conceptnet_rocks.conceptnet5.uri import conjunction_uri


def transform_for_linked_data(edge):
    """
    Modify an edge (assertion) in place to contain values that are appropriate
    for a Linked Data API.

    Although this code isn't actually responsible for what an API returns
    (see the conceptnet-web repository for that), it helps to deal with what
    edge dictionaries should contain here.

    The relevant changes are:

    - Remove the 'features' list
    - Rename 'uri' to '@id'
    - Make 'start', 'end', and 'rel' into dictionaries with an '@id' and
      'label', removing the separate 'surfaceStart' and 'surfaceEnd'
      attributes
    - All dictionaries should have an '@id'. For the edge itself, it's the
      URI. Without this, we get RDF blank nodes, which are awful.
    - Set '@type' on objects representing edges and sources. (Nodes get their
      @type from the `ld_node` function.)
    """
    for source in edge['sources']:
        conj = conjunction_uri(*sorted(source.values()))
        source['@id'] = conj
        source['@type'] = 'Source'
    edge['@id'] = edge['uri']
    del edge['uri']
    edge['@type'] = 'Edge'

    start_uri = edge['start']
    end_uri = edge['end']
    rel_uri = edge['rel']
    start_label = edge.pop('surfaceStart', None)
    end_label = edge.pop('surfaceEnd', None)
    edge['start'] = ld_node(start_uri, start_label)
    edge['end'] = ld_node(end_uri, end_label)
    edge['rel'] = ld_node(rel_uri, None)
    if 'other' in edge:
        # TODO: Find out when we use this, or remove it if we don't use it
        if edge['other'] == start_uri:
            edge['other'] = edge['start']
        elif edge['other'] == end_uri:
            edge['other'] = edge['end']
        else:
            edge['rel'] = ld_node(rel_uri, None)

    if edge.get('surfaceText') is None:
        edge['surfaceText'] = None

    edge['weight'] = float(edge['weight'])

    return edge





from arango import ArangoClient
from conceptnet_rocks.conceptnet5.edges import transform_for_linked_data
from graph_garden import arangodb
from hashlib import sha1
from pathlib import Path
from tqdm import tqdm
from typing import Optional, Any, Dict, List
import dask.dataframe as dd
import orjson
import os
import subprocess


DEFAULT_DATABASE = "conceptnet"


def convert_uri_to_ascii(uri: str) -> str:
    return sha1(uri.encode()).hexdigest()


def get_node_id(uri_ascii: str) -> str:
    return f"nodes/{uri_ascii}"


def line_count(file_path):
    f = open(file_path, "rb")
    result = 0
    buf_size = 1024 * 1024
    read_f = f.raw.read

    buf = read_f(buf_size)
    while buf:
        result += buf.count(b"\n")
        buf = read_f(buf_size)

    return result


def load_dump_into_database(
        dump_path: Path,
        edge_count: Optional[int] = None,
        connection_uri: str = arangodb.DEFAULT_CONNECTION_URI,
        database: str = DEFAULT_DATABASE,
        root_password: str = arangodb.DEFAULT_ROOT_PASSWORD,
        arangodb_exe_path: Path = arangodb.DEFAULT_INSTALL_PATH,
        data_path: Path = arangodb.DEFAULT_DATA_PATH,
):
    with arangodb.instance(
        connection_uri=connection_uri,
        root_password=root_password,
        arangodb_exe_path=arangodb_exe_path,
        data_path=data_path,
    ):
        client = ArangoClient(hosts=connection_uri)

        root_credentials = {
            "username": "root",
            "password": root_password,
        }

        sys_db = client.db(arangodb.SYSTEM_DATABASE, **root_credentials)
        if sys_db.has_database(database):
            raise RuntimeError(f"Database {database} already exist")

        sys_db.create_database(database)

        db = client.db(database, **root_credentials)

        graph = db.create_graph("conceptnet")

        graph.create_vertex_collection("nodes")

        graph.create_edge_definition(
            edge_collection="edges",
            from_vertex_collections=["nodes"],
            to_vertex_collections=["nodes"]
        )

        db.create_collection("sources")

        dump_path = Path(dump_path).expanduser().resolve()

        df = dd.read_csv(dump_path, sep="\t", header=None)
        batch_size = 2**10

        if edge_count is None:
            edge_count = line_count(dump_path)

        node_list = []
        edge_list = []
        source_list = []

        nodes_path = data_path / "nodes.jsonl"
        edges_path = data_path / "edges.jsonl"
        sources_path = data_path / "sources.jsonl"

        if not nodes_path.is_file() or not edges_path.is_file() or not sources_path.is_file():
            with open(nodes_path, "wb") as nodes_file, \
                    open(edges_path, "wb") as edges_file, \
                    open(sources_path, "wb") as sources_file:
                nodes_keys = set()

                for row in tqdm(df.itertuples(), unit=" edges", initial=1, total=edge_count):
                    i, assertion_uri, relation_uri, start_uri, end_uri, edge_data = row

                    start_uri += "/"
                    end_uri += "/"

                    edge_data_dict = orjson.loads(edge_data)
                    edge_data_dict["dataset"] += "/"

                    def source_with_trailing_slash(source: Dict[str, str]) -> Dict[str, str]:
                        result = {}
                        for field in ["activity", "contributor", "process"]:
                            if field in source:
                                result[field] = source[field] + "/"
                        return result

                    edge_data_dict["sources"] = [
                        source_with_trailing_slash(source)
                        for source in edge_data_dict["sources"]
                    ]

                    start_node_key = convert_uri_to_ascii(start_uri)
                    if start_node_key not in nodes_keys:
                        nodes_keys.add(start_node_key)
                        node_list.append({"_key": start_node_key, "uri": start_uri})

                    end_node_key = convert_uri_to_ascii(end_uri)
                    if end_node_key not in nodes_keys:
                        nodes_keys.add(end_node_key)
                        node_list.append({"_key": end_node_key, "uri": end_uri})

                    start_node_id = get_node_id(start_node_key)
                    end_node_id = get_node_id(end_node_key)

                    edge_key = convert_uri_to_ascii(assertion_uri)

                    edge_list.append({
                        "_key": edge_key,
                        "_from": start_node_id,
                        "_to": end_node_id,
                        "uri": assertion_uri,
                        "rel": relation_uri,
                        **edge_data_dict,
                    })

                    for source in edge_data_dict["sources"]:
                        source_list.append({
                            "edge_key": edge_key,
                            **source,
                        })

                    finished = i == edge_count - 1
                    if i % batch_size == 0 or finished:
                        def write_to_file(f, item_list: List):
                            for item in item_list:
                                f.write(orjson.dumps(item))
                                f.write(b"\n")
                            item_list.clear()

                        write_to_file(f=nodes_file, item_list=node_list)
                        write_to_file(f=edges_file, item_list=edge_list)
                        write_to_file(f=sources_file, item_list=source_list)
                        if finished:
                            break

        arangoimport_path = arangodb.get_exe_path(path=arangodb_exe_path, program_name="arangoimport")
        for file_path, collection in [(nodes_path, "nodes"), (edges_path, "edges"), (sources_path, "sources")]:
            subprocess.run([
                arangoimport_path,
                "--threads", f"{os.cpu_count()}",
                "--file", file_path,
                "--type", "jsonl",
                "--server.database", database,
                "--server.endpoint", connection_uri.replace("http", "tcp"),
                "--server.username", "root",
                "--server.password", root_password,
                "--collection", collection,
                "--on-duplicate", "ignore",
                "--skip-validation",
                "--log.level", "error",
            ])
            file_path.unlink()

        nodes = db.collection("nodes")
        edges = db.collection("edges")
        sources = db.collection("sources")

        print("Creating nodes.uri index")
        nodes.add_persistent_index(["uri"], unique=True)
        print("Creating edges.dataset index")
        edges.add_persistent_index(["dataset"])
        print("Creating edges.rel index")
        edges.add_persistent_index(["rel"])
        print("Creating edges.uri index")
        edges.add_persistent_index(["uri"], unique=True)
        print("Creating sources.activity index")
        sources.add_persistent_index(["activity"])
        print("Creating sources.contributor index")
        sources.add_persistent_index(["contributor"])
        print("Creating sources.process index")
        sources.add_persistent_index(["process"])


class AssertionFinder:
    def __init__(
            self,
            database: str = DEFAULT_DATABASE,
            connection_uri: str = arangodb.DEFAULT_CONNECTION_URI,
            root_password: str = arangodb.DEFAULT_ROOT_PASSWORD,
            arangodb_exe_path: Path = arangodb.DEFAULT_INSTALL_PATH,
            data_path: Path = arangodb.DEFAULT_DATA_PATH,
            close_stdout_and_stderr: bool = False,
    ):
        self._arangodb_arbiter = arangodb.start_if_not_running(
            connection_uri=connection_uri,
            root_password=root_password,
            arangodb_exe_path=arangodb_exe_path,
            data_path=data_path,
            close_stdout_and_stderr=close_stdout_and_stderr,
        )
        self._client = ArangoClient(hosts=connection_uri)
        self._db = self._client.db(database, username="root", password=root_password)

    def __del__(self):
        arangodb.stop_arbiter(self._arangodb_arbiter)

    def lookup(self, uri: str, limit: int = 100, offset: int = 0):
        # noinspection PyShadowingNames
        def perform_query(query: str, query_vars: Dict[str, Any]) -> List[Dict[str, Any]]:
            return [
                transform_for_linked_data(data)
                for data in self._db.aql.execute(query, bind_vars=query_vars)
            ]

        limit_and_merge = """
            limit @offset, @limit
            let trimmed_sources = (
              for source in edge.sources
                return merge(
                  has(source, "activity") ? {activity: rtrim(source.activity, "/")} : {},
                  has(source, "contributor") ? {contributor: rtrim(source.contributor, "/")} : {},
                  has(source, "process") ? {process: rtrim(source.process, "/")} : {}
                )
            )
            return distinct merge({
              start: rtrim(document(edge._from).uri, "/"),
              end: rtrim(document(edge._to).uri, "/"),
              dataset: rtrim(edge.dataset, "/"),
              sources: trimmed_sources,
            }, unset(edge, "_from", "_to", "_key", "_id", "_rev", "dataset", "sources"))
        """

        if uri.startswith("/c/") or uri.startswith("http"):
            query = f"""
                for start_node in nodes
                  filter start_node.uri >= concat(@uri, "/") and start_node.uri < concat(@uri, "0")
                  for node, edge in any start_node edges
                  {limit_and_merge}
            """
            query_vars = {"limit": limit, "offset": offset, "uri": uri}
            return perform_query(query=query, query_vars=query_vars)
        elif uri.startswith("/r/"):
            query = f"""
                for edge in edges
                  filter edge.rel == @rel
                  {limit_and_merge}
            """
            query_vars = {"limit": limit, "offset": offset, "rel": uri}
            return perform_query(query=query, query_vars=query_vars)
        elif uri.startswith("/s/"):
            query = f"""
                for s in sources
                  filter (
                    (s.activity >= concat(@source, "/") and s.activity < concat(@source, "0"))
                    or
                    (s.contributor >= concat(@source, "/") and s.contributor < concat(@source, "0"))
                    or
                    (s.process >= concat(@source, "/") and s.process < concat(@source, "0"))
                  )
                  for edge in edges
                    filter s.edge_key == edge._key
                    {limit_and_merge}
            """
            query_vars = {"limit": limit, "offset": offset, "source": uri}
            return perform_query(query=query, query_vars=query_vars)
        elif uri.startswith("/d/"):
            query = f"""
                for edge in edges
                  filter edge.dataset >= concat(@dataset, "/") and edge.dataset < concat(@dataset, "0")
                  {limit_and_merge}
            """
            query_vars = {"limit": limit, "offset": offset, "dataset": uri}
            return perform_query(query=query, query_vars=query_vars)
        elif uri.startswith("/a/"):
            query = f"""
                for edge in edges
                  filter edge.uri == @uri
                  {limit_and_merge}
            """
            query_vars = {"limit": limit, "offset": offset, "uri": uri}
            return perform_query(query=query, query_vars=query_vars)
        else:
            raise ValueError(f"{uri} isn't a supported")

    def clear_cache(self) -> None:
        self._db.aql.cache.clear()



from conceptnet_rocks.conceptnet5.uri import uri_to_label, is_term, split_uri, get_uri_language, uri_prefix
from urllib.parse import urlparse


def ld_node(uri, label=None):
    """
    Convert a ConceptNet URI into a dictionary suitable for Linked Data.
    """
    if label is None:
        label = uri_to_label(uri)
    ld = {'@id': uri, 'label': label}
    if is_term(uri):
        pieces = split_uri(uri)
        ld['language'] = get_uri_language(uri)

        # Get a reasonably-distinct sense label for the term.
        # Usually it will be the part of speech, but when we have fine-grained
        # information from Wikipedia or WordNet, it'll include the last
        # component as well.
        if len(pieces) > 3:
            ld['sense_label'] = pieces[3]

        if len(pieces) > 4 and pieces[4] in ('wp', 'wn'):
            ld['sense_label'] += ', ' + pieces[-1]

        ld['term'] = uri_prefix(uri)
        ld['@type'] = 'Node'
    elif uri.startswith('http'):
        domain = urlparse(uri).netloc
        ld['site'] = domain
        ld['term'] = uri

        # OpenCyc is down and UMBEL doesn't host their vocabulary on the
        # Web. This property indicates whether you can follow a link
        # via HTTP and retrieve more information.
        ld['site_available'] = True
        if domain in {'sw.opencyc.org', 'umbel.org', 'wikidata.dbpedia.org'}:
            ld['site_available'] = False
        ld['path'] = urlparse(uri).path
        ld['@type'] = 'Node'
    elif uri.startswith('/r/'):
        ld['@type'] = 'Relation'
    return ld



from conceptnet_rocks.database import load_dump_into_database
from graph_garden import arangodb
from pathlib import Path
from typing import Optional
import typer


app = typer.Typer()


@app.command()
def load(
        dump_path: Path,
        edge_count: Optional[int] = None,
        arangodb_exe_path: Path = arangodb.DEFAULT_INSTALL_PATH,
        data_path: Path = arangodb.DEFAULT_DATA_PATH,
):
    load_dump_into_database(
        dump_path=dump_path,
        edge_count=edge_count,
        arangodb_exe_path=arangodb_exe_path,
        data_path=data_path,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['conceptnet_rocks', 'conceptnet_rocks.conceptnet5']

package_data = \
{'': ['*']}

install_requires = \
['graph-garden>=0.1.2,<0.2.0',
 'orjson>=3.3.1,<4.0.0',
 'python-arango>=6.0.0,<7.0.0',
 'tqdm>=4.48.2,<5.0.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['conceptnet-rocks-load = conceptnet_rocks.cli:app']}

setup_kwargs = {
    'name': 'conceptnet-rocks',
    'version': '0.1.1',
    'description': 'Python library to work with ConceptNet offline',
    'long_description': '# ConceptNet Rocks!\n\nWork is in progress.\n\nThe library comes with Apache License 2.0, and is separate from ConceptNet itself, although it uses some parts of its code. The ConceptNet is available under [CC-BY-SA-4.0](https://creativecommons.org/licenses/by-sa/4.0/) license. See [here](https://github.com/commonsense/conceptnet5/wiki/Copying-and-sharing-ConceptNet) for the list of conditions for using ConceptNet data.\n\nThis is the official citation for ConceptNet if you use it in research:\n\n> Robyn Speer, Joshua Chin, and Catherine Havasi. 2017. "ConceptNet 5.5: An Open Multilingual Graph of General Knowledge." In proceedings of AAAI 31.\n\n## Installation\n\n```bash\npip install conceptnet-rocks\n```\n\n## Usage\n\n### Install ArangoDB\n\nConceptNet Rocks uses [ArangoDB](https://www.arangodb.com/) for storage, managed by a companion Python [Graph Garden library](https://github.com/ldtoolkit/graph-garden/)  that is automatically installed with ConceptNet Rocks.\n\nGraph Garden can manage the ArangoDB installation for you. To download the latest version of ArangoDB from official website and install it to `~/.arangodb` folder, simply run:\n\n```bash\ngraph-garden arangodb install\n```\n\nFor more options execute:\n\n```bash\ngraph-garden arangodb install --help\n```\n\n### Load CSV dump into database\n\nThen you need to load the ConceptNet CSV dump into database. The dump can be downloaded from https://github.com/commonsense/conceptnet5/wiki/Downloads\n\nLet\'s assume you\'ve downloaded the dump to `~/conceptnet-data/assertions.csv`.\n\nTo load the dump, execute:\n```bash\nconceptnet-rocks-load ~/conceptnet-data/assertions.csv\n```\n\nThis command will create database in `~/.arangodb/data`. For more options execute:\n\n```bash\nconceptnet-rocks-load --help\n```\n\n### Run queries\n\nNow you can query ConceptNet. ConceptNet Rocks uses the same simple API as ConceptNet5 for querying:\n\n```python\nfrom conceptnet_rocks import AssertionFinder\n\naf = AssertionFinder()\nprint(af.lookup("/c/en/test"))\nprint(af.lookup("/r/Antonym"))\nprint(af.lookup("/s/process/wikiparsec/2"))\nprint(af.lookup("/d/wiktionary/en"))\nprint(af.lookup("/a/[/r/Antonym/,/c/ang/gecyndelic/a/,/c/ang/ungecynde/]"))\n```\n\nConceptNet Rocks uses the same JSON-LD format as the original ConceptNet5:\n```python\nfrom conceptnet_rocks import AssertionFinder\nfrom pprint import pprint\n\naf = AssertionFinder()\npprint(af.lookup("/c/en/blow_dryer"))\n# [\n# ...\n#  {\'@id\': \'/a/[/r/AtLocation/,/c/en/blow_dryer/,/c/en/beauty_salon/]\',\n#   \'@type\': \'Edge\',\n#   \'dataset\': \'/d/conceptnet/4/en\',\n#   \'end\': {\'@id\': \'/c/en/beauty_salon\',\n#           \'@type\': \'Node\',\n#           \'label\': \'a beauty salon\',\n#           \'language\': \'en\',\n#           \'term\': \'/c/en/beauty_salon\'},\n#   \'license\': \'cc:by/4.0\',\n#   \'rel\': {\'@id\': \'/r/AtLocation\', \'@type\': \'Relation\', \'label\': \'AtLocation\'},\n#   \'sources\': [{\'@id\': \'/and/[/s/activity/omcs/omcs1_possibly_free_text/,/s/contributor/omcs/bedume/]\',\n#                \'@type\': \'Source\',\n#                \'activity\': \'/s/activity/omcs/omcs1_possibly_free_text\',\n#                \'contributor\': \'/s/contributor/omcs/bedume\'}],\n#   \'start\': {\'@id\': \'/c/en/blow_dryer\',\n#             \'@type\': \'Node\',\n#             \'label\': \'a blow dryer\',\n#             \'language\': \'en\',\n#             \'term\': \'/c/en/blow_dryer\'},\n#   \'surfaceText\': \'You are likely to find [[a blow dryer]] in [[a beauty salon]]\',\n#   \'weight\': 1.0}\n# ...\n# ]\n```\n\n## FAQ\n\n### Why did you create yet another library if original ConceptNet5 exists?\n\n1. Performance. Our benchmark (https://github.com/ldtoolkit/conceptnet-benchmark) has shown that ConceptNet Rocks is\nalmost 5 times faster than ConceptNet5 for querying assertions by concepts.\n2. The original ConceptNet5 library requires PostgreSQL. PostgreSQL does not support the graph databases as a primary model, while ArangoDB is a multi-model database for graph.\n3. PostgreSQL generally requires either root permissions to install it using a package manager, or the compilation step. Not anyone have root permissions on their machine or have the compiler installed. ConceptNet Rocks library uses ArangoDB, which can be installed without root permissions using simple command.\n\n### Why is the library called ConceptNet Rocks?\n\n1. Under the hood ArangoDB uses key-value storage called RocksDB.\n2. Performance-wise, it does rock!\n',
    'author': 'Roman Inflianskas',
    'author_email': 'infroma@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ldtoolkit/conceptnet-rocks',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
